#extension GL_EXT_gpu_shader4 : enable
#extension GL_ARB_gpu_shader5 : enable
varying vec2 texcoord;

// YCoCg Color Space Conversion Function
vec4 texelFetchYC(sampler2D tex, ivec2 uv, int level)
{
    vec4 c = texelFetch(tex, uv, level);
#ifdef USE_YCOCG
    // Inline YCoCg conversion to avoid extra temporary variables.
    float y = c.r * 0.25 + c.g * 0.5 + c.b * 0.25;
    float co = c.r * 0.5 - c.b * 0.5;
    float cg = -c.r * 0.25 + c.g * 0.5 - c.b * 0.25;
    return vec4(y, co, cg, c.a);
#else
    return c;
#endif
}

// Fetch history pixel
vec4 fetchHistoryPx(ivec2 pix)
{
    return vec4(texelFetchYC(colortex4, pix, 0).xyz, 1.0);
}

// Cubic Hermite interpolation
vec3 CubicHermite(vec3 A, vec3 B, vec3 C, vec3 D, float t)
{
    float t2 = t * t;
    float t3 = t2 * t;
    return ((-0.5 * A + 1.5 * B - 1.5 * C + 0.5 * D) * t3) + ((A - 2.5 * B + 2.0 * C - 0.5 * D) * t2) + ((-0.5 * A + 0.5 * C) * t) + B;
}

vec3 OptimizedCubicTextureSample(vec2 P)
{
    // Convert normalized coordinate to pixel space.
    vec2 pixel = P * viewResolution + 0.5;
    vec2 frac = fract(pixel);
    ivec2 ipixel = ivec2(pixel) - 1;

    // Manually unroll the 3-row interpolation.
    vec3 row0;
    {
        vec3 C0 = fetchHistoryPx(ipixel + ivec2(-1, -1)).rgb;
        vec3 C1 = fetchHistoryPx(ipixel + ivec2(0, -1)).rgb;
        vec3 C2 = fetchHistoryPx(ipixel + ivec2(1, -1)).rgb;
        row0 = CubicHermite(C0, C1, C2, C2, frac.x);
    }

    vec3 row1;
    {
        vec3 C0 = fetchHistoryPx(ipixel + ivec2(-1, 0)).rgb;
        vec3 C1 = fetchHistoryPx(ipixel + ivec2(0, 0)).rgb;
        vec3 C2 = fetchHistoryPx(ipixel + ivec2(1, 0)).rgb;
        row1 = CubicHermite(C0, C1, C2, C2, frac.x);
    }

    vec3 row2;
    {
        vec3 C0 = fetchHistoryPx(ipixel + ivec2(-1, 1)).rgb;
        vec3 C1 = fetchHistoryPx(ipixel + ivec2(0, 1)).rgb;
        vec3 C2 = fetchHistoryPx(ipixel + ivec2(1, 1)).rgb;
        row2 = CubicHermite(C0, C1, C2, C2, frac.x);
    }

    return CubicHermite(row0, row1, row2, row2, frac.y);
}

/*
float MinDepth3x3_old(sampler2D tex, ivec2 fragCoord)
{
    float d0 = texelFetch(tex, fragCoord, 0).x;
    float d1 = texelFetch(tex, fragCoord + ivec2(-1, -1), 0).x;
    float d2 = texelFetch(tex, fragCoord + ivec2(0, -1), 0).x;
    float d3 = texelFetch(tex, fragCoord + ivec2(1, -1), 0).x;
    float d4 = texelFetch(tex, fragCoord + ivec2(-1, 0), 0).x;
    float d5 = texelFetch(tex, fragCoord + ivec2(1, 0), 0).x;
    float d6 = texelFetch(tex, fragCoord + ivec2(-1, 1), 0).x;
    float d7 = texelFetch(tex, fragCoord + ivec2(0, 1), 0).x;
    float d8 = texelFetch(tex, fragCoord + ivec2(1, 1), 0).x;
    float minDepth = d0;
    minDepth = min(minDepth, d1);
    minDepth = min(minDepth, d2);
    minDepth = min(minDepth, d3);
    minDepth = min(minDepth, d4);
    minDepth = min(minDepth, d5);
    minDepth = min(minDepth, d6);
    minDepth = min(minDepth, d7);
    minDepth = min(minDepth, d8);
    return minDepth;
}
*/

// Unrolled 3x3 minimum depth computation to reduce loop overhead
float MinDepth3x3(sampler2D tex, ivec2 fragCoord)
{
    // Convert fragCoord to normalized texture coordinates
    vec2 uv = (vec2(fragCoord) + 0.5) / vec2(textureSize(tex, 0));

    // Gather four depth values from one corner of the 3x3 neighborhood
    vec4 depths = textureGather(tex, uv, 0);

    // Compute the minimum among the gathered depths
    float minDepth = min(min(depths.x, depths.y), min(depths.z, depths.w));

    // Note: This doesn't cover all 9 pixels, but if the depth is fairly coherent,
    // this approximation may be acceptable.
    return minDepth;
}

#ifdef DISTANT_HORIZONS
// Unrolled 3x3 minimum depth computation for distant horizons
float MinDepth3x3DH(sampler2D tex, ivec2 fragCoord)
{
    float d0 = texelFetch(tex, fragCoord, 0).x;
    float d1 = texelFetch(tex, fragCoord + ivec2(-1, -1), 0).x;
    float d2 = texelFetch(tex, fragCoord + ivec2(0, -1), 0).x;
    float d3 = texelFetch(tex, fragCoord + ivec2(1, -1), 0).x;
    float d4 = texelFetch(tex, fragCoord + ivec2(-1, 0), 0).x;
    float d5 = texelFetch(tex, fragCoord + ivec2(1, 0), 0).x;
    float d6 = texelFetch(tex, fragCoord + ivec2(-1, 1), 0).x;
    float d7 = texelFetch(tex, fragCoord + ivec2(0, 1), 0).x;
    float d8 = texelFetch(tex, fragCoord + ivec2(1, 1), 0).x;
    float minDepth = d0;
    minDepth = min(minDepth, d1);
    minDepth = min(minDepth, d2);
    minDepth = min(minDepth, d3);
    minDepth = min(minDepth, d4);
    minDepth = min(minDepth, d5);
    minDepth = min(minDepth, d6);
    minDepth = min(minDepth, d7);
    minDepth = min(minDepth, d8);
    return minDepth;
}
vec3 ReprojectDH2(vec3 position)
{
    bool isHand = position.z < 0.6;
    if (position.z < 1.0 && !isHand)
    {
        position = ViewSpaceFromScreenSpace(position, dhProjectionInverse);
        position = mat3(gbufferModelViewInverse) * position + gbufferModelViewInverse[3].xyz;
        position += (cameraPosition - previousCameraPosition);
        position = mat3(gbufferPreviousModelView) * position + gbufferPreviousModelView[3].xyz;
        position = ScreenSpaceFromViewSpace(position, dhPreviousProjection);
    }
    else
    {
        position = ViewSpaceFromScreenSpace(position, dhProjectionInverse);
        position = mat3(gbufferModelViewInverse) * position + gbufferModelViewInverse[3].xyz;
        position = mat3(gbufferPreviousModelView) * position + gbufferPreviousModelView[3].xyz;
        position = ScreenSpaceFromViewSpace(position, dhPreviousProjection);
    }
    return position;
}

#endif

vec3 Reproject(vec3 position)
{
    bool isHand = position.z < 0.6;
    position = ViewSpaceFromScreenSpace(position, gbufferProjectionInverse);

    vec3 worldPosition = mat3(gbufferModelViewInverse) * position + gbufferModelViewInverse[3].xyz;
    worldPosition += (!isHand && position.z < 1.0) ? (cameraPosition - previousCameraPosition) : vec3(0.0);

    vec3 previousViewPos = mat3(gbufferPreviousModelView) * worldPosition + gbufferPreviousModelView[3].xyz;
    return ScreenSpaceFromViewSpace(previousViewPos, gbufferPreviousProjection);
}

// Simple clipping function for color history
vec3 clip_aabb_simple(vec3 min_color, vec3 max_color, vec3 history_color)
{
    return clamp(history_color, min_color, max_color);
}

/* RENDERTARGETS: 4 */
void main()
{
    vec3 color = vec3(0.0);
    vec2 uv = texcoord;
    ivec2 texel = ivec2(uv * viewResolution);

#ifdef TAA
    float z = texture(depthtex0, uv).x;
    float alpha = texture(colortex1, uv).a;
    float entity = texture(colortex5, uv).g;
    float offcenter_rej = TAA_OFFCENTER_REJECTION;
    float historyWeight = TAA_HISTORY_WEIGHT;
    float depth = MinDepth3x3(depthtex0, texel);
    vec3 current = texture2D(colortex0, uv + offsets[framemod8] * texelSize * 0.5).rgb;

    vec3 position = vec3(uv, depth);

    #ifndef WATER_PBR
    if (alpha > 0.1)
    {
        float depth2 = texelFetch(depthtex1, texel, 0).x;

        vec3 position2 = vec3(uv, depth2);

        position = mix(position2, position, pow(alpha, 10.0));
    }
    #endif

    vec3 reprojectedPosition = Reproject(position);

    #ifdef DISTANT_HORIZONS
    if (z >= 1.0)
    {
        depth = MinDepth3x3DH(dhDepthTex0, texel);
        position = vec3(uv, depth);
        reprojectedPosition = ReprojectDH2(position);

        historyWeight = 0.98;
        offcenter_rej = 0.1;
    }
    #endif

    vec3 history = OptimizedCubicTextureSample(reprojectedPosition.xy);

    if (any(lessThan(reprojectedPosition.xy, vec2(0.0))) ||
        any(greaterThan(reprojectedPosition.xy, vec2(1.0))))
    {
        historyWeight = 0.0;
    }

    #ifdef TAA_CLIP
    vec2 pcd = 1.0 - abs(2.0 * fract(reprojectedPosition.xy * viewResolution) - 1.0);
    historyWeight *= sqrt(pcd.x * pcd.y) * offcenter_rej + (1.0 - offcenter_rej);

    // Consolidated neighborhood min/max using on-the-fly texture lookups.
    vec3 minC = current, maxC = current;
    vec2 baseCoord = (vec2(texel) + 0.5) / viewResolution;
    vec2 invRes = 1.0 / viewResolution;

    // Process each neighbor by offset, reusing the computed values.
    minC = min(minC, texture(colortex0, baseCoord + vec2(-1, -1) * invRes).rgb);
    maxC = max(maxC, texture(colortex0, baseCoord + vec2(-1, -1) * invRes).rgb);

    minC = min(minC, texture(colortex0, baseCoord + vec2(0, -1) * invRes).rgb);
    maxC = max(maxC, texture(colortex0, baseCoord + vec2(0, -1) * invRes).rgb);

    minC = min(minC, texture(colortex0, baseCoord + vec2(1, -1) * invRes).rgb);
    maxC = max(maxC, texture(colortex0, baseCoord + vec2(1, -1) * invRes).rgb);

    minC = min(minC, texture(colortex0, baseCoord + vec2(-1, 0) * invRes).rgb);
    maxC = max(maxC, texture(colortex0, baseCoord + vec2(-1, 0) * invRes).rgb);

    minC = min(minC, texture(colortex0, baseCoord + vec2(1, 0) * invRes).rgb);
    maxC = max(maxC, texture(colortex0, baseCoord + vec2(1, 0) * invRes).rgb);

    minC = min(minC, texture(colortex0, baseCoord + vec2(-1, 1) * invRes).rgb);
    maxC = max(maxC, texture(colortex0, baseCoord + vec2(-1, 1) * invRes).rgb);

    minC = min(minC, texture(colortex0, baseCoord + vec2(0, 1) * invRes).rgb);
    maxC = max(maxC, texture(colortex0, baseCoord + vec2(0, 1) * invRes).rgb);

    minC = min(minC, texture(colortex0, baseCoord + vec2(1, 1) * invRes).rgb);
    maxC = max(maxC, texture(colortex0, baseCoord + vec2(1, 1) * invRes).rgb);

    history = clip_aabb_simple(minC, maxC, history);
    #endif

    #ifdef NO_ENTITY_TAA
    if (entity > 0.1)
    {
        historyWeight = clamp(1.0 - entity, 0.15, historyWeight);
    }
    #endif

    color = mix(current, history, historyWeight);
#else
    color = texture(colortex0, uv).rgb;
#endif

    gl_FragData[0].rgb = clamp(color, 0.0001, 1000.0);
    gl_FragData[0].a = texture(colortex4, uv).a;
}
